from vulkan import *
import glfw
from glfw.GLFW import GLFW_NO_API, GLFW_CLIENT_API, GLFW_RESIZABLE, GLFW_FALSE

import array
import numpy as np
from PIL import Image

reqExtensions = [VK_KHR_SWAPCHAIN_EXTENSION_NAME, ]


# vkGetInstanceProcAddr ----------------------------------------------------------------------------
class InstanceProcAddr:
    T = None

    def __init__(self, func):
        self.F = func

    def __call__(self, *args, **kwargs):
        FName = self.F.__name__
        F = InstanceProcAddr.procFunc(FName)
        if F:
            return F(*args, **kwargs)
        else:
            return VK_ERROR_EXTENSION_NOT_PRESENT

    @staticmethod
    def procFunc(funcName):
        return vkGetInstanceProcAddr(InstanceProcAddr.T, funcName)


# vkGetDeviceProcAddr ------------------------------------------------------------------------------
class DeviceProcAddr(InstanceProcAddr):
    @staticmethod
    def procFunc(funcName):
        return vkGetDeviceProcAddr(InstanceProcAddr.T, funcName)


# vkGetInstanceProcAddr ----------------------------------------------------------------------------
@InstanceProcAddr
def vkDestroySurfaceKHR(instance, surface, pAllocator):
    pass


@InstanceProcAddr
def vkGetPhysicalDeviceSurfaceSupportKHR(physicalDevice, queueFamilyIndex, surface):
    pass


@InstanceProcAddr
def vkGetPhysicalDeviceSurfaceCapabilitiesKHR(physicalDevice, surface):
    pass


@InstanceProcAddr
def vkGetPhysicalDeviceSurfaceFormatsKHR(physicalDevice, surface):
    pass


@InstanceProcAddr
def vkGetPhysicalDeviceSurfacePresentModesKHR(physicalDevice, surface):
    pass


# vkGetDeviceProcAddr ------------------------------------------------------------------------------
@DeviceProcAddr
def vkCreateSwapchainKHR(logicalDevice, pCreateInfo, pAllocator):
    pass


@DeviceProcAddr
def vkDestroySwapchainKHR(logicalDevice, swapchain, pAllocator):
    pass


@DeviceProcAddr
def vkGetSwapchainImagesKHR(logicalDevice, swapchain):
    pass


@DeviceProcAddr
def vkAcquireNextImageKHR(logicalDevice, swapchain, timeout, semaphore, fence):
    pass


@DeviceProcAddr
def vkQueuePresentKHR(queue, pPresentInfo):
    pass


# trieda QueueFamilyIndices ------------------------------------------------------------------------
class QueueFamilyIndices:
    def __init__(self):
        self.graphicsFamily = None  # graphics queue family index
        self.presentFamily = None  # present queue family index

    def getState(self):
        return (self.graphicsFamily is not None) and (self.presentFamily is not None)

    def getQueues(self):
        return [self.graphicsFamily, self.presentFamily]


# trieda SwapchainSupportDetails -------------------------------------------------------------------
class SwapchainSupportDetails:
    def __init__(self):
        self.capabilities = None
        self.formats = []
        self.presentModes = []


# trieda pre pracu s vrcholmi ----------------------------------------------------------------------
class Vertex:
    def __init__(self):
        self.pos = []
        self.color = []
        self.texcoord = []

    POS = array.array('f', [0.0, 0.0, 0.0])
    COLOR = array.array('f', [0.0, 0.0, 0.0])
    TEXCOORD = array.array('f', [0.0, 0.0])

    @staticmethod
    def getVertexBindingDescription():
        bindingDescription = VkVertexInputBindingDescription(
            binding=0,
            stride=len(Vertex.POS) * Vertex.POS.itemsize + len(Vertex.COLOR) * Vertex.COLOR.itemsize + len(
                Vertex.TEXCOORD) * Vertex.TEXCOORD.itemsize,
            inputRate=VK_VERTEX_INPUT_RATE_VERTEX
        )
        return bindingDescription

    @staticmethod
    def getVertexAttributeDescriptions():
        pos = VkVertexInputAttributeDescription(
            location=0,
            binding=0,
            format=VK_FORMAT_R32G32B32_SFLOAT,
            offset=0
        )
        color = VkVertexInputAttributeDescription(
            location=1,
            binding=0,
            format=VK_FORMAT_R32G32B32_SFLOAT,
            offset=len(Vertex.POS) * Vertex.POS.itemsize,
        )
        texcoord = VkVertexInputAttributeDescription(
            location=2,
            binding=0,
            format=VK_FORMAT_R32G32_SFLOAT,
            offset=len(Vertex.POS) * Vertex.POS.itemsize + len(Vertex.COLOR) * Vertex.COLOR.itemsize,
        )
        return [pos, color, texcoord]


# trieda pre pracu s UBO - Uniform Buffer Object ---------------------------------------------------
class UniformBufferObject:
    def __init__(self):
        self.model = np.identity(4, np.float32)
        self.view = np.identity(4, np.float32)
        self.proj = np.identity(4, np.float32)

    def toArray(self):
        return np.concatenate((self.model, self.view, self.proj))

    @property
    def nbytes(self):
        return self.proj.nbytes + self.view.nbytes + self.model.nbytes


# trieda VulkanApp ---------------------------------------------------------------------------------
class VulkanApp:
    # konstruktor
    def __init__(self):
        self.width = 800                      # sirka okna
        self.height = 600                     # vyska okna
        self.window = None                    # GLFW okno
        self.instance = None                  # izolovana instancia API Vulkan
        self.surface = None                   # WIN surface VkSurfaceKHR
        self.physicalDevice = VK_NULL_HANDLE  # Fyzicke zariadenie
        self.logicalDevice = None             # Logicke zariadenie
        self.graphicsQueue = None             # Graphics Queue
        self.presentQueue = None              # Present Queue
        self.swapchain = None                 # Swapchain
        self.swapchainImages = None           # Swapchain Images
        self.swapchainImageFormat = None      # Swapchain ImageFormat
        self.swapchainExtent = None           # Swapchain Extent
        self.swapchainImageViews = []         # Swapchain ImageViews
        self.renderPass = None                # RenderPass
        self.swapchainFramebuffers = []       # Swapchain Framebuffers
        self.descriptorSetLayout = None       # DescriptorSetLayout
        self.descriptorPool = None            # DescriptorPool
        self.descriptorSet = None             # DescriptorSet
        self.ubo = UniformBufferObject()      # UniformBufferObject
        self.uniformBuffer = None             # UniformBuffer
        self.uniformBufferMemory = None       # UniformBufferMemory
        self.pipelineLayout = None            # GraphicsPipelineLayout
        self.graphicsPipeline = None          # GraphicsPipeline
        self.commandPool = None               # Command Pool
        self.commandBuffers = None            # Command Buffers
        self.imageAvailableSemaphore = None   # Semaphore
        self.renderFinishedSemaphore = None   # Semaphore

        self.vertices = None                  # Vrcholy
        self.vertexBuffer = None              # Zasobnik vrcholov
        self.vertexBufferMemory = None        # Pamat zasobnika vrcholov

        self.textureImage = None              # Textura
        self.textureImageMemory = None
        self.textureImageView = None
        self.textureSampler = None

    # Vytvorenie GLFW okna -------------------------------------------------------------------------
    def createWindow(self):
        glfw.init()
        glfw.window_hint(GLFW_CLIENT_API, GLFW_NO_API)
        glfw.window_hint(GLFW_RESIZABLE, GLFW_FALSE)

        self.window = glfw.create_window(self.width, self.height, "PyGLSL v2.1", None, None)

    # Ziskanie indexov graphicsFamily a presentFamily Queues ---------------------------------------
    def getQueueFamilyIndices(self, device):
        indices = QueueFamilyIndices()  # instancia triedy QueueFamilyIndices
        queueFamilies = vkGetPhysicalDeviceQueueFamilyProperties(device)
        i = 0
        for q in queueFamilies:
            if q.queueFlags & VK_QUEUE_GRAPHICS_BIT:
                indices.graphicsFamily = i
            if vkGetPhysicalDeviceSurfaceSupportKHR(device, i, self.surface):
                indices.presentFamily = i
            i += 1
        return indices

    # Ziskanie detailov podpory Swapchain ----------------------------------------------------------
    def getSwapchainSupportDetails(self, device):
        details = SwapchainSupportDetails()  # instancia triedy SwapchainSupportDetails
        details.capabilities = vkGetPhysicalDeviceSurfaceCapabilitiesKHR(device, self.surface)
        details.formats = vkGetPhysicalDeviceSurfaceFormatsKHR(device, self.surface)
        details.presentModes = vkGetPhysicalDeviceSurfacePresentModesKHR(device, self.surface)
        return details

    # Kontrola pouzitelnosti fyzickeho zariadenia --------------------------------------------------
    def isPhysicalDeviceSuitable(self, device):
        # Ziskanie indexov graphicsFamily a presentFamily Queues
        indices = self.getQueueFamilyIndices(device)
        supExtensions = vkEnumerateDeviceExtensionProperties(device, None)
        for e in reqExtensions:
            extSupported = False
            for f in supExtensions:
                if e == f.extensionName:
                    extSupported = True
            if not extSupported:
                return False
        # Ziskanie detailov podpory Swapchain
        swapchainDetails = self.getSwapchainSupportDetails(device)
        return indices.getState() and \
               len(swapchainDetails.formats) != 0 and \
               len(swapchainDetails.presentModes) != 0

    # Vytvorenie Shader modulu ---------------------------------------------------------------------
    def createShaderModule(self, fpath):
        with open(fpath, 'rb') as file:
            code = file.read()
            ShaderModuleCreateInfo = VkShaderModuleCreateInfo(
                sType=VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO,
                codeSize=len(code),
                pCode=code
            )
            return vkCreateShaderModule(self.logicalDevice, ShaderModuleCreateInfo, None)

    # Priprava 3D objektov -------------------------------------------------------------------------
    def prepare3Dobjects(self):
        self.vertices = array.array('f', [
            # pos            color           texcoord
            -0.5, -0.5, 0.0, 0.0, 0.25, 0.0, 0.0, 1.0,
             0.5,  0.5, 0.0, 0.0, 0.25, 0.0, 1.0, 0.0,
            -0.5,  0.5, 0.0, 0.0, 0.25, 0.0, 0.0, 0.0,
            -0.5, -0.5, 0.0, 0.0, 0.5,  0.0, 0.0, 1.0,
             0.5, -0.5, 0.0, 0.0, 0.5,  0.0, 1.0, 1.0,
             0.5,  0.5, 0.0, 0.0, 0.5,  0.0, 1.0, 0.0,
        ])
        return

    # Vytvorenie zasobnika -------------------------------------------------------------------------
    def createBuffer(self, size, usage, properties):
        BufferCreateInfo = VkBufferCreateInfo(
            size=size,
            usage=usage,
            sharingMode=VK_SHARING_MODE_EXCLUSIVE
        )
        buffer = vkCreateBuffer(self.logicalDevice, BufferCreateInfo, None)

        BufferMemoryRequirements = vkGetBufferMemoryRequirements(self.logicalDevice, buffer)
        PhysicalDeviceMemoryProperties = vkGetPhysicalDeviceMemoryProperties(self.physicalDevice)
        memTypeIndex = -1
        for i, p in enumerate(PhysicalDeviceMemoryProperties.memoryTypes):
            if (BufferMemoryRequirements.memoryTypeBits & (1 << i)) and ((p.propertyFlags & properties) == properties):
                memTypeIndex = i
                break

        MemoryAllocateInfo = VkMemoryAllocateInfo(
            allocationSize=BufferMemoryRequirements.size,
            memoryTypeIndex=memTypeIndex
        )
        bufferMemory = vkAllocateMemory(self.logicalDevice, MemoryAllocateInfo, None)

        vkBindBufferMemory(self.logicalDevice, buffer, bufferMemory, 0)

        return buffer, bufferMemory

    # Vytvorenie Image -----------------------------------------------------------------------------
    def createImage(self, W, H, F, tiling, usage, properties):
        ImageCreateInfo = VkImageCreateInfo(
            imageType=VK_IMAGE_TYPE_2D,
            extent=[W, H, 1],
            mipLevels=1,
            arrayLayers=1,
            format=F,
            samples=VK_SAMPLE_COUNT_1_BIT,
            tiling=tiling,
            usage=usage,
            sharingMode=VK_SHARING_MODE_EXCLUSIVE,
            initialLayout=VK_IMAGE_LAYOUT_UNDEFINED
        )
        image = vkCreateImage(self.logicalDevice, ImageCreateInfo, None)

        ImageMemoryRequirements = vkGetImageMemoryRequirements(self.logicalDevice, image)
        PhysicalDeviceMemoryProperties = vkGetPhysicalDeviceMemoryProperties(self.physicalDevice)
        memTypeIndex = -1
        for i, p in enumerate(PhysicalDeviceMemoryProperties.memoryTypes):
            if (ImageMemoryRequirements.memoryTypeBits & (1 << i)) and ((p.propertyFlags & properties) == properties):
                memTypeIndex = i
                break
        MemoryAllocateInfo = VkMemoryAllocateInfo(
            allocationSize=ImageMemoryRequirements.size,
            memoryTypeIndex=memTypeIndex
        )

        imageMemory = vkAllocateMemory(self.logicalDevice, MemoryAllocateInfo, None)

        vkBindImageMemory(self.logicalDevice, image, imageMemory, 0)

        return image, imageMemory

    def beginSingleTimeCommands(self):
        CommandBufferAllocateInfo = VkCommandBufferAllocateInfo(
            level=VK_COMMAND_BUFFER_LEVEL_PRIMARY,
            commandPool=self.commandPool,
            commandBufferCount=1
        )
        commandBuffer = vkAllocateCommandBuffers(self.logicalDevice, CommandBufferAllocateInfo)[0]
        beginInfo = VkCommandBufferBeginInfo(flags=VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT)
        vkBeginCommandBuffer(commandBuffer, beginInfo)

        return commandBuffer

    def endSingleTimeCommands(self, commandBuffer):
        vkEndCommandBuffer(commandBuffer)

        submitInfo = VkSubmitInfo(pCommandBuffers=[commandBuffer])

        vkQueueSubmit(self.graphicsQueue, 1, [submitInfo], VK_NULL_HANDLE)
        vkQueueWaitIdle(self.graphicsQueue)

        vkFreeCommandBuffers(self.logicalDevice, self.commandPool, 1, [commandBuffer])

    def transitionImageLayout(self, image, oldLayout, newLayout):
        cmdBuffer = self.beginSingleTimeCommands()

        ImageSubresourceRange = VkImageSubresourceRange(
            aspectMask=VK_IMAGE_ASPECT_COLOR_BIT,
            baseMipLevel=0,
            levelCount=1,
            baseArrayLayer=0,
            layerCount=1
        )
        ImageMemoryBarrier = VkImageMemoryBarrier(
            oldLayout=oldLayout,
            newLayout=newLayout,
            srcQueueFamilyIndex=VK_QUEUE_FAMILY_IGNORED,
            dstQueueFamilyIndex=VK_QUEUE_FAMILY_IGNORED,
            image=image,
            subresourceRange=ImageSubresourceRange
        )
        if oldLayout == VK_IMAGE_LAYOUT_UNDEFINED and newLayout == VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL:
            ImageMemoryBarrier.srcAccessMask = 0
            ImageMemoryBarrier.dstAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT

            sourceStage = VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT
            destinationStage = VK_PIPELINE_STAGE_TRANSFER_BIT
        elif oldLayout == VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL and newLayout == \
                VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL:
            ImageMemoryBarrier.srcAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT
            ImageMemoryBarrier.dstAccessMask = VK_ACCESS_SHADER_READ_BIT

            sourceStage = VK_PIPELINE_STAGE_TRANSFER_BIT
            destinationStage = VK_PIPELINE_STAGE_FRAGMENT_SHADER_BIT
        else:
            raise Exception('unsupported layout transition!')

        vkCmdPipelineBarrier(cmdBuffer,
                             sourceStage,
                             destinationStage,
                             0,
                             0, None,
                             0, None,
                             1, ImageMemoryBarrier)

        self.endSingleTimeCommands(cmdBuffer)

    def copyBufferToImage(self, buffer, image, W, H):
        cmdBuffer = self.beginSingleTimeCommands()

        ImageSubresourceLayers = VkImageSubresourceLayers(
            aspectMask=VK_IMAGE_ASPECT_COLOR_BIT,
            mipLevel=0,
            baseArrayLayer=0,
            layerCount=1
        )
        BufferImageCopy = VkBufferImageCopy(
            bufferOffset=0,
            bufferRowLength=0,
            bufferImageHeight=0,
            imageSubresource=ImageSubresourceLayers,
            imageOffset=[0, 0],
            imageExtent=[W, H, 1]
        )

        vkCmdCopyBufferToImage(cmdBuffer, buffer, image, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, 1, BufferImageCopy)

        self.endSingleTimeCommands(cmdBuffer)

    # Inicializacia API Vulkan ---------------------------------------------------------------------
    def initVulkan(self):
        # Vytvorenie izolovanej instancie API Vulkan -----------------------------------------------
        # ApplicationInfo
        ApplicationInfo = VkApplicationInfo(
            sType=VK_STRUCTURE_TYPE_APPLICATION_INFO,
            pApplicationName="PyGLSL v2.1",
            applicationVersion=VK_MAKE_VERSION(2, 1, 0),
            pEngineName="N/A",
            engineVersion=VK_MAKE_VERSION(1, 0, 0),
            apiVersion=VK_API_VERSION_1_0)
        # Vulkan instance extensions potrebne pre GLFW
        Extensions = glfw.get_required_instance_extensions()
        # InstanceCreateInfo
        InstanceCreateInfo = VkInstanceCreateInfo(
            sType=VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO,
            pApplicationInfo=ApplicationInfo,
            enabledExtensionCount=len(Extensions),
            ppEnabledExtensionNames=Extensions,
            enabledLayerCount=0,
            ppEnabledLayerNames=None)
        try:
            self.instance = vkCreateInstance(InstanceCreateInfo, None)
            InstanceProcAddr.T = self.instance
        except Exception:
            raise RuntimeError("Chyba vytvorenia instancie API Vulkan!")

        # Vytvorenie WIN surface -------------------------------------------------------------------
        surface_ptr = ffi.new('VkSurfaceKHR[1]')
        glfw.create_window_surface(self.instance, self.window, None, surface_ptr)
        self.surface = surface_ptr[0]
        if self.surface is None:
            raise RuntimeError("Chyba vytvorenia WIN Surface!")

        # Vyber fyzickeho zariadenia - GPU ---------------------------------------------------------
        devices = vkEnumeratePhysicalDevices(self.instance)
        for d in devices:
            if self.isPhysicalDeviceSuitable(d):  # Kontrola pouzitelnosti fyzickeho zariadenia
                self.physicalDevice = d
                break
        if self.physicalDevice == VK_NULL_HANDLE:
            raise RuntimeError("Nemozno vybrat pouzitelne fyzicke zariadenie - GPU!")
        GPUindices = self.getQueueFamilyIndices(self.physicalDevice)

        # Vytvorenie logickeho zariadenia ----------------------------------------------------------
        DeviceQueueCreateInfo = []
        for q in GPUindices.getQueues():
            DeviceQueueCreateInfo.append(VkDeviceQueueCreateInfo(
                sType=VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO,
                queueFamilyIndex=q,
                queueCount=1,
                pQueuePriorities=[1, ])
            )
        PhysicalDeviceFeatures = VkPhysicalDeviceFeatures()
        DeviceCreateInfo = VkDeviceCreateInfo(
            sType=VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO,
            flags=0,
            pQueueCreateInfos=DeviceQueueCreateInfo,
            queueCreateInfoCount=len(DeviceQueueCreateInfo),
            pEnabledFeatures=PhysicalDeviceFeatures,
            enabledExtensionCount=len(reqExtensions),
            ppEnabledExtensionNames=reqExtensions,
            enabledLayerCount=0,
            ppEnabledLayerNames=None
        )
        try:
            self.logicalDevice = vkCreateDevice(self.physicalDevice, DeviceCreateInfo, None)
        except Exception:
            raise RuntimeError("Chyba vytvorenia logickeho zariadenia!")
        # Graphics a Present Queues
        self.graphicsQueue = vkGetDeviceQueue(self.logicalDevice, GPUindices.graphicsFamily, 0)
        self.presentQueue = vkGetDeviceQueue(self.logicalDevice, GPUindices.presentFamily, 0)

        # Vytvorenie Swapchain ---------------------------------------------------------------------
        swapchainDetails = self.getSwapchainSupportDetails(self.physicalDevice)
        imgCount = swapchainDetails.capabilities.minImageCount + 1
        if (swapchainDetails.capabilities.maxImageCount > 0) and \
                (imgCount > swapchainDetails.capabilities.maxImageCount):
            imgCount = swapchainDetails.capabilities.maxImageCount
        surfaceFormat = swapchainDetails.formats[0]
        for f in swapchainDetails.formats:
            if f.format == VK_FORMAT_B8G8R8A8_SRGB and f.colorSpace == VK_COLOR_SPACE_SRGB_NONLINEAR_KHR:
                surfaceFormat = f
                break
        presentMode = VK_PRESENT_MODE_FIFO_KHR  # double-buffering
        for p in swapchainDetails.presentModes:
            if p == VK_PRESENT_MODE_MAILBOX_KHR:
                presentMode = p  # triple-buffering
                break
        w = max(swapchainDetails.capabilities.minImageExtent.width,
                min(swapchainDetails.capabilities.maxImageExtent.width, self.width))
        h = max(swapchainDetails.capabilities.minImageExtent.height,
                min(swapchainDetails.capabilities.maxImageExtent.height, self.height))
        surfaceExtent = VkExtent2D(w, h)
        if GPUindices.graphicsFamily != GPUindices.presentFamily:
            cI_imageSharingMode = VK_SHARING_MODE_CONCURRENT
            cI_queueFamilyIndexCount = 2
            cI_pQueueFamilyIndices = [GPUindices.graphicsFamily, GPUindices.presentFamily]
        else:
            cI_imageSharingMode = VK_SHARING_MODE_EXCLUSIVE
            cI_queueFamilyIndexCount = 0
            cI_pQueueFamilyIndices = None
        SwapchainCreateInfo = VkSwapchainCreateInfoKHR(
            sType=VK_STRUCTURE_TYPE_SWAPCHAIN_CREATE_INFO_KHR,
            surface=self.surface,
            minImageCount=imgCount,
            imageFormat=surfaceFormat.format,
            imageColorSpace=surfaceFormat.colorSpace,
            imageExtent=surfaceExtent,
            imageArrayLayers=1,
            imageUsage=VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT,
            imageSharingMode=cI_imageSharingMode,
            queueFamilyIndexCount=cI_queueFamilyIndexCount,
            pQueueFamilyIndices=cI_pQueueFamilyIndices,
            preTransform=swapchainDetails.capabilities.currentTransform,
            compositeAlpha=VK_COMPOSITE_ALPHA_OPAQUE_BIT_KHR,
            presentMode=presentMode,
            clipped=VK_TRUE,
            oldSwapchain=VK_NULL_HANDLE
        )
        self.swapchain = vkCreateSwapchainKHR(self.logicalDevice, SwapchainCreateInfo, None)
        self.swapchainImages = vkGetSwapchainImagesKHR(self.logicalDevice, self.swapchain)
        self.swapchainImageFormat = surfaceFormat.format
        self.swapchainExtent = surfaceExtent

        # Vytvorenie ImageViews --------------------------------------------------------------------
        for i in self.swapchainImages:
            ImageViewCreateInfo = VkImageViewCreateInfo(
                sType=VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO,
                image=i,
                viewType=VK_IMAGE_VIEW_TYPE_2D,
                format=self.swapchainImageFormat,
                components=VkComponentMapping(r=VK_COMPONENT_SWIZZLE_IDENTITY, g=VK_COMPONENT_SWIZZLE_IDENTITY,
                                              b=VK_COMPONENT_SWIZZLE_IDENTITY, a=VK_COMPONENT_SWIZZLE_IDENTITY),
                subresourceRange=VkImageSubresourceRange(aspectMask=VK_IMAGE_ASPECT_COLOR_BIT,
                                                         baseMipLevel=0, levelCount=1,
                                                         baseArrayLayer=0, layerCount=1)
            )
            self.swapchainImageViews.append(vkCreateImageView(self.logicalDevice, ImageViewCreateInfo, None))

        # Vytvorenie RederPass ---------------------------------------------------------------------
        colorAttachment = VkAttachmentDescription(
            format=self.swapchainImageFormat,
            samples=VK_SAMPLE_COUNT_1_BIT,
            loadOp=VK_ATTACHMENT_LOAD_OP_CLEAR,
            storeOp=VK_ATTACHMENT_STORE_OP_STORE,
            stencilLoadOp=VK_ATTACHMENT_LOAD_OP_DONT_CARE,
            stencilStoreOp=VK_ATTACHMENT_STORE_OP_DONT_CARE,
            initialLayout=VK_IMAGE_LAYOUT_UNDEFINED,
            finalLayout=VK_IMAGE_LAYOUT_PRESENT_SRC_KHR
        )
        colorAttachmentRef = VkAttachmentReference(
            attachment=0,
            layout=VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL
        )
        subpassDescription = VkSubpassDescription(
            pipelineBindPoint=VK_PIPELINE_BIND_POINT_GRAPHICS,
            colorAttachmentCount=1,
            pColorAttachments=colorAttachmentRef
        )
        subpassDependency = VkSubpassDependency(
            srcSubpass=VK_SUBPASS_EXTERNAL,
            dstSubpass=0,
            srcStageMask=VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT,
            srcAccessMask=0,
            dstStageMask=VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT,
            dstAccessMask=VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT
        )
        RenderPassCreateInfo = VkRenderPassCreateInfo(
            sType=VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO,
            attachmentCount=1,
            pAttachments=colorAttachment,
            subpassCount=1,
            pSubpasses=subpassDescription,
            dependencyCount=1,
            pDependencies=subpassDependency
        )
        self.renderPass = vkCreateRenderPass(self.logicalDevice, RenderPassCreateInfo, None)

        # Vytvorenie Framebuffers ------------------------------------------------------------------
        for v in self.swapchainImageViews:
            attachments = [v, ]
            FramebufferCreateInfo = VkFramebufferCreateInfo(
                sType=VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO,
                renderPass=self.renderPass,
                attachmentCount=1,
                pAttachments=attachments,
                width=self.swapchainExtent.width,
                height=self.swapchainExtent.height,
                layers=1
            )
            self.swapchainFramebuffers.append(vkCreateFramebuffer(self.logicalDevice, FramebufferCreateInfo, None))

        # Vytvorenie Shader modulov ----------------------------------------------------------------
        VSModule = self.createShaderModule("shaders/vert.spv")
        VSPipelineShaderStageCreateInfo = VkPipelineShaderStageCreateInfo(
            sType=VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO,
            stage=VK_SHADER_STAGE_VERTEX_BIT,
            module=VSModule,
            pName="main"
        )
        FSModule = self.createShaderModule("shaders/frag.spv")
        FSPipelineShaderStageCreateInfo = VkPipelineShaderStageCreateInfo(
            sType=VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO,
            stage=VK_SHADER_STAGE_FRAGMENT_BIT,
            module=FSModule,
            pName="main"
        )
        shaderStages = [VSPipelineShaderStageCreateInfo, FSPipelineShaderStageCreateInfo]

        # Vytvorenie DescriptorSetLayout -----------------------------------------------------------
        uboLayoutBinding = VkDescriptorSetLayoutBinding(
            binding=0,
            descriptorType=VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER,
            descriptorCount=1,
            stageFlags=VK_SHADER_STAGE_VERTEX_BIT
        )
        samplerLayoutBinding = VkDescriptorSetLayoutBinding(
            binding=1,
            descriptorType=VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER,
            descriptorCount=1,
            stageFlags=VK_SHADER_STAGE_FRAGMENT_BIT
        )
        DescriptorSetLayoutInfo = VkDescriptorSetLayoutCreateInfo(
            pBindings=[uboLayoutBinding, samplerLayoutBinding]
        )
        self.descriptorSetLayout = vkCreateDescriptorSetLayout(self.logicalDevice, DescriptorSetLayoutInfo, None)

        # Vytvorenie GraphicsPipeline --------------------------------------------------------------
        # VS binding
        VertexBindingDescription = Vertex.getVertexBindingDescription()
        VertexAttributeDescription = Vertex.getVertexAttributeDescriptions()
        PipelineVertexInputStateCreateInfo = VkPipelineVertexInputStateCreateInfo(
            sType=VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_STATE_CREATE_INFO,
            # vertexBindingDescriptionCount=0,
            pVertexBindingDescriptions=[VertexBindingDescription],
            # vertexAttributeDescriptionCount=0
            pVertexAttributeDescriptions=VertexAttributeDescription
        )
        # assembly
        PipelineInputAssemblyStateCreateInfo = VkPipelineInputAssemblyStateCreateInfo(
            sType=VK_STRUCTURE_TYPE_PIPELINE_INPUT_ASSEMBLY_STATE_CREATE_INFO,
            topology=VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST,
            primitiveRestartEnable=VK_FALSE
        )
        # viewport & scissor
        viewport = VkViewport(
            x=0,
            y=0,
            width=self.swapchainExtent.width,
            height=self.swapchainExtent.height,
            minDepth=0.0,
            maxDepth=1.0
        )
        scissor = VkRect2D(
            offset=[0, 0],
            extent=self.swapchainExtent
        )
        PipelineViewportStateCreateInfo = VkPipelineViewportStateCreateInfo(
            sType=VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_STATE_CREATE_INFO,
            viewportCount=1,
            pViewports=viewport,
            scissorCount=1,
            pScissors=scissor
        )
        # rasterizer
        PipelineRasterizationStateCreateInfo = VkPipelineRasterizationStateCreateInfo(
            sType=VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_CREATE_INFO,
            depthClampEnable=VK_FALSE,
            rasterizerDiscardEnable=VK_FALSE,
            polygonMode=VK_POLYGON_MODE_FILL,
            lineWidth=1.0,
            cullMode=VK_CULL_MODE_BACK_BIT,
            frontFace=VK_FRONT_FACE_CLOCKWISE,
            depthBiasEnable=VK_FALSE
        )
        # multisampling
        PipelineMultisampleStateCreateInfo = VkPipelineMultisampleStateCreateInfo(
            sType=VK_STRUCTURE_TYPE_PIPELINE_MULTISAMPLE_STATE_CREATE_INFO,
            sampleShadingEnable=VK_FALSE,
            rasterizationSamples=VK_SAMPLE_COUNT_1_BIT
        )
        # color blending
        colorBlendAttachment = VkPipelineColorBlendAttachmentState(
            colorWriteMask=VK_COLOR_COMPONENT_R_BIT | VK_COLOR_COMPONENT_G_BIT | VK_COLOR_COMPONENT_B_BIT |
                           VK_COLOR_COMPONENT_A_BIT,
            blendEnable=VK_FALSE
        )
        PipelineColorBlendStateCreateInfo = VkPipelineColorBlendStateCreateInfo(
            sType=VK_STRUCTURE_TYPE_PIPELINE_COLOR_BLEND_STATE_CREATE_INFO,
            logicOpEnable=VK_FALSE,
            attachmentCount=1,
            pAttachments=colorBlendAttachment,
            blendConstants=[0.0, 0.0, 0.0, 0.0]
        )
        PipelineLayoutCreateInfo = VkPipelineLayoutCreateInfo(
            # sType=VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO,
            pushConstantRangeCount=0,
            pSetLayouts=[self.descriptorSetLayout]
        )
        self.pipelineLayout = vkCreatePipelineLayout(self.logicalDevice, PipelineLayoutCreateInfo, None)
        GraphicsPipelineCreateInfo = VkGraphicsPipelineCreateInfo(
            sType=VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_CREATE_INFO,
            stageCount=2,
            pStages=shaderStages,
            pVertexInputState=PipelineVertexInputStateCreateInfo,
            pInputAssemblyState=PipelineInputAssemblyStateCreateInfo,
            pViewportState=PipelineViewportStateCreateInfo,
            pRasterizationState=PipelineRasterizationStateCreateInfo,
            pMultisampleState=PipelineMultisampleStateCreateInfo,
            pDepthStencilState=None,
            pColorBlendState=PipelineColorBlendStateCreateInfo,
            layout=self.pipelineLayout,
            renderPass=self.renderPass,
            subpass=0
        )
        self.graphicsPipeline = \
            vkCreateGraphicsPipelines(self.logicalDevice, VK_NULL_HANDLE, 1, GraphicsPipelineCreateInfo, None)[0]

        # Zrusenie Shader modulov ------------------------------------------------------------------
        vkDestroyShaderModule(self.logicalDevice, VSModule, None)
        vkDestroyShaderModule(self.logicalDevice, FSModule, None)

        # Vytvorenie vertexBuffer ------------------------------------------------------------------
        BufferCreateInfo = VkBufferCreateInfo(
            size=len(self.vertices) * self.vertices.itemsize,
            usage=VK_BUFFER_USAGE_VERTEX_BUFFER_BIT,
            sharingMode=VK_SHARING_MODE_EXCLUSIVE
        )
        self.vertexBuffer = vkCreateBuffer(self.logicalDevice, BufferCreateInfo, None)

        BufferMemoryRequirements = vkGetBufferMemoryRequirements(self.logicalDevice, self.vertexBuffer)
        PhysicalDeviceMemoryProperties = vkGetPhysicalDeviceMemoryProperties(self.physicalDevice)
        memTypeIndex = -1
        for i, p in enumerate(PhysicalDeviceMemoryProperties.memoryTypes):
            if (BufferMemoryRequirements.memoryTypeBits & (1 << i)) and \
                    ((p.propertyFlags & VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT) ==
                     VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT):
                memTypeIndex = i
                break

        MemoryAllocateInfo = VkMemoryAllocateInfo(
            allocationSize=BufferMemoryRequirements.size,
            memoryTypeIndex=memTypeIndex
        )
        self.vertexBufferMemory = vkAllocateMemory(self.logicalDevice, MemoryAllocateInfo, None)

        vkBindBufferMemory(self.logicalDevice, self.vertexBuffer, self.vertexBufferMemory, 0)

        data = vkMapMemory(self.logicalDevice, self.vertexBufferMemory, 0, BufferCreateInfo.size, 0)
        vertexPtr = ffi.cast('float *', self.vertices.buffer_info()[0])
        ffi.memmove(data, vertexPtr, BufferCreateInfo.size)
        vkUnmapMemory(self.logicalDevice, self.vertexBufferMemory)

        # Vytvorenie CommandPool -------------------------------------------------------------------
        CommandPoolCreateInfo = VkCommandPoolCreateInfo(
            sType=VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO,
            queueFamilyIndex=GPUindices.graphicsFamily
        )
        self.commandPool = vkCreateCommandPool(self.logicalDevice, CommandPoolCreateInfo, None)

        # Nahratie textury -------------------------------------------------------------------------
        tex = Image.open('textures/default.png')
        tex.putalpha(1)
        tex_width = tex.width
        tex_height = tex.height
        tex_size = tex_width * tex_height * 4

        stagingBuffer, stagingMem = self.createBuffer(tex_size, VK_BUFFER_USAGE_TRANSFER_SRC_BIT,
                                                      VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT |
                                                      VK_MEMORY_PROPERTY_HOST_COHERENT_BIT)

        data = vkMapMemory(self.logicalDevice, stagingMem, 0, tex_size, 0)
        ffi.memmove(data, tex.tobytes(), tex_size)
        vkUnmapMemory(self.logicalDevice, stagingMem)

        del tex

        self.textureImage, self.textureImageMemory = self.createImage(tex_width, tex_height,
                                                                      VK_FORMAT_R8G8B8A8_UNORM,
                                                                      VK_IMAGE_TILING_OPTIMAL,
                                                                      VK_IMAGE_USAGE_TRANSFER_DST_BIT |
                                                                      VK_IMAGE_USAGE_SAMPLED_BIT,
                                                                      VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT)

        self.transitionImageLayout(self.textureImage, VK_IMAGE_LAYOUT_UNDEFINED,
                                   VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL)
        self.copyBufferToImage(stagingBuffer, self.textureImage, tex_width, tex_height)
        self.transitionImageLayout(self.textureImage, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
                                   VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL)

        vkDestroyBuffer(self.logicalDevice, stagingBuffer, None)
        vkFreeMemory(self.logicalDevice, stagingMem, None)

        ssr = VkImageSubresourceRange(VK_IMAGE_ASPECT_COLOR_BIT,0, 1, 0, 1)
        ImageViewCreateInfo = VkImageViewCreateInfo(
            image=self.textureImage,
            viewType=VK_IMAGE_VIEW_TYPE_2D,
            format=VK_FORMAT_R8G8B8A8_UNORM,
            subresourceRange=ssr
        )
        self.textureImageView = vkCreateImageView(self.logicalDevice, ImageViewCreateInfo, None)

        SamplerCreateInfo = VkSamplerCreateInfo(
            magFilter=VK_FILTER_LINEAR,
            minFilter=VK_FILTER_LINEAR,
            addressModeU=VK_SAMPLER_ADDRESS_MODE_REPEAT,
            addressModeV=VK_SAMPLER_ADDRESS_MODE_REPEAT,
            addressModeW=VK_SAMPLER_ADDRESS_MODE_REPEAT,
            anisotropyEnable=True,
            maxAnisotropy=16,
            compareEnable=False,
            compareOp=VK_COMPARE_OP_ALWAYS,
            borderColor=VK_BORDER_COLOR_INT_OPAQUE_BLACK,
            unnormalizedCoordinates=False
        )
        self.textureSampler = vkCreateSampler(self.logicalDevice, SamplerCreateInfo, None)

        # Vytvorenie DescriptorPool ----------------------------------------------------------------
        poolSize1 = VkDescriptorPoolSize(
            type=VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER,
            descriptorCount=1
        )
        poolSize2 = VkDescriptorPoolSize(
            type=VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER,
            descriptorCount=1
        )
        DescriptorPoolCreateInfo = VkDescriptorPoolCreateInfo(
            pPoolSizes=[poolSize1, poolSize2],
            maxSets=1
        )
        self.descriptorPool = vkCreateDescriptorPool(self.logicalDevice, DescriptorPoolCreateInfo, None)

        # Vytvorenie DescriptorSet -----------------------------------------------------------------
        self.uniformBuffer, self.uniformBufferMemory = self.createBuffer(self.ubo.nbytes,
                                                                         VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT,
                                                                         VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT |
                                                                         VK_MEMORY_PROPERTY_HOST_COHERENT_BIT)
        layouts = [self.descriptorSetLayout]
        DescriptorSetAllocateInfo = VkDescriptorSetAllocateInfo(
            descriptorPool=self.descriptorPool,
            pSetLayouts=layouts
        )
        self.descriptorSet = vkAllocateDescriptorSets(self.logicalDevice, DescriptorSetAllocateInfo)
        DescriptorBufferInfo = VkDescriptorBufferInfo(
            buffer=self.uniformBuffer,
            offset=0,
            range=self.ubo.nbytes
        )
        DescriptorImageInfo = VkDescriptorImageInfo(
            imageLayout=VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL,
            imageView=self.textureImageView,
            sampler=self.textureSampler
        )
        descriptWrite1 = VkWriteDescriptorSet(
            dstSet=self.descriptorSet[0],
            dstBinding=0,
            dstArrayElement=0,
            descriptorType=VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER,
            pBufferInfo=[DescriptorBufferInfo]
        )
        descriptWrite2 = VkWriteDescriptorSet(
            dstSet=self.descriptorSet[0],
            dstBinding=1,
            dstArrayElement=0,
            descriptorType=VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER,
            pImageInfo=[DescriptorImageInfo]
        )
        vkUpdateDescriptorSets(self.logicalDevice, 2, [descriptWrite1, descriptWrite2], 0, None)

        # Alokovanie CommandBuffers ----------------------------------------------------------------
        CommandBufferAllocateInfo = VkCommandBufferAllocateInfo(
            sType=VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO,
            commandPool=self.commandPool,
            level=VK_COMMAND_BUFFER_LEVEL_PRIMARY,
            commandBufferCount=len(self.swapchainFramebuffers)
        )
        commandBuffers = vkAllocateCommandBuffers(self.logicalDevice, CommandBufferAllocateInfo)
        self.commandBuffers = [ffi.addressof(commandBuffers, i)[0] for i in range(len(self.swapchainFramebuffers))]
        for i, b in enumerate(self.commandBuffers):
            CommandBufferBeginInfo = VkCommandBufferBeginInfo(
                sType=VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO,
                flags=VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT
            )
            RenderPassBeginInfo = VkRenderPassBeginInfo(
                sType=VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO,
                renderPass=self.renderPass,
                framebuffer=self.swapchainFramebuffers[i],
                renderArea=[[0, 0], self.swapchainExtent]
            )
            clearColor = VkClearValue([[0.2, 0.2, 0.2, 1.0]])
            RenderPassBeginInfo.clearValueCount = 1
            RenderPassBeginInfo.pClearValues = ffi.addressof(clearColor)
            vkBeginCommandBuffer(b, CommandBufferBeginInfo)                               # BEGIN Command Buffer
            vkCmdBeginRenderPass(b, RenderPassBeginInfo, VK_SUBPASS_CONTENTS_INLINE)       # BEGIN Render Pass
            vkCmdBindPipeline(b, VK_PIPELINE_BIND_POINT_GRAPHICS, self.graphicsPipeline)    # BIND Graphics Pipeline
            vkCmdBindVertexBuffers(b, 0, 1, [self.vertexBuffer], [0])                       # BIND Vertex Buffer
            vkCmdBindDescriptorSets(b, VK_PIPELINE_BIND_POINT_GRAPHICS, self.pipelineLayout, 0, 1,
                                    self.descriptorSet, 0, None)                            # BIND Descriptor Sets
            vkCmdDraw(b, 6, 1, 0, 0)                                                        # DRAW
            vkCmdEndRenderPass(b)                                                          # END Render Pass
            vkEndCommandBuffer(b)                                                         # END Command Buffer

        # Vytvorenie Semaphore ---------------------------------------------------------------------
        SemaphoreCreateInfo = VkSemaphoreCreateInfo(
            sType=VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO
        )
        self.imageAvailableSemaphore = vkCreateSemaphore(self.logicalDevice, SemaphoreCreateInfo, None)
        self.renderFinishedSemaphore = vkCreateSemaphore(self.logicalDevice, SemaphoreCreateInfo, None)

    # Vykreslenie snimky ---------------------------------------------------------------------------
    def drawFrame(self):
        imageIndex = vkAcquireNextImageKHR(self.logicalDevice, self.swapchain, 18446744073709551615,
                                           self.imageAvailableSemaphore, VK_NULL_HANDLE)
        SubmitInfo = VkSubmitInfo(
            sType=VK_STRUCTURE_TYPE_SUBMIT_INFO,
            waitSemaphoreCount=1,
            pWaitSemaphores=[self.imageAvailableSemaphore, ],
            pWaitDstStageMask=[VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT, ],
            commandBufferCount=1,
            pCommandBuffers=[self.commandBuffers[imageIndex], ],
            signalSemaphoreCount=1,
            pSignalSemaphores=[self.renderFinishedSemaphore, ]
        )
        vkQueueSubmit(self.graphicsQueue, 1, SubmitInfo, VK_NULL_HANDLE)
        swapChains = [self.swapchain, ]
        PresentInfoKHR = VkPresentInfoKHR(
            sType=VK_STRUCTURE_TYPE_PRESENT_INFO_KHR,
            waitSemaphoreCount=1,
            pWaitSemaphores=[self.renderFinishedSemaphore, ],
            swapchainCount=1,
            pSwapchains=swapChains,
            pImageIndices=[imageIndex, ]
        )
        vkQueuePresentKHR(self.presentQueue, PresentInfoKHR)
        vkQueueWaitIdle(self.presentQueue)

    # Ukoncenie aplikacie --------------------------------------------------------------------------
    def exit(self):
        # Zrusenie Image
        vkDestroySampler(self.logicalDevice, self.textureSampler, None)
        vkDestroyImageView(self.logicalDevice, self.textureImageView, None)
        vkDestroyImage(self.logicalDevice, self.textureImage, None)
        vkFreeMemory(self.logicalDevice, self.textureImageMemory, None)
        # Zrusenie DescriptorSet
        vkDestroyDescriptorSetLayout(self.logicalDevice, self.descriptorSetLayout, None)
        vkDestroyDescriptorPool(self.logicalDevice, self.descriptorPool, None)
        # Zrusenie VertexBuffer
        vkDestroyBuffer(self.logicalDevice, self.vertexBuffer, None)
        vkFreeMemory(self.logicalDevice, self.vertexBufferMemory, None)
        # Zrusenie Semaphore
        vkDestroySemaphore(self.logicalDevice, self.imageAvailableSemaphore, None)
        vkDestroySemaphore(self.logicalDevice, self.renderFinishedSemaphore, None)
        # Zrusenie CommandPool
        vkDestroyCommandPool(self.logicalDevice, self.commandPool, None)
        # Zrusenie GraphicsPipeline
        vkDestroyPipeline(self.logicalDevice, self.graphicsPipeline, None)
        # Zrusenie GraphicsPipelineLayout
        vkDestroyPipelineLayout(self.logicalDevice, self.pipelineLayout, None)
        # Zrusenie Framebuffers
        for f in self.swapchainFramebuffers:
            vkDestroyFramebuffer(self.logicalDevice, f, None)
        # Zrusenie RenderPass
        vkDestroyRenderPass(self.logicalDevice, self.renderPass, None)
        # Zrusenie ImageView
        for i in self.swapchainImageViews:
            vkDestroyImageView(self.logicalDevice, i, None)
        # Zrusenie Swapchain
        vkDestroySwapchainKHR(self.logicalDevice, self.swapchain, None)
        # Zrusenie logickeho zariadenia
        vkDestroyDevice(self.logicalDevice, None)
        # Zrusenie WIN surface
        vkDestroySurfaceKHR(self.instance, self.surface, None)
        # Zrusenie instancie
        vkDestroyInstance(self.instance, None)
        # Zrusenie GLFW okna
        glfw.destroy_window(self.window)
        # END
        glfw.terminate()

    # Hlavny programovy cyklus ---------------------------------------------------------------------
    def mainLoop(self):
        while not glfw.window_should_close(self.window):
            glfw.poll_events()  # zachytenie udalosti
            self.drawFrame()    # vykreslenie snimky
        vkDeviceWaitIdle(self.logicalDevice)


# Instancia triedy VulkanApp
vApp = VulkanApp()

# Vytvorenie GLFW okna
vApp.createWindow()
# Priprava 3D objektov
vApp.prepare3Dobjects()
# Inicializacia API Vulkan
vApp.initVulkan()
# Hlavny programovy cyklus
vApp.mainLoop()
# Ukoncenie aplikacie
vApp.exit()
